//********************************************************************************************************************************
//*                                                                                                                              *
//*     Thumbnails Database Management unit  3delite 2006-2009                                                                  *
//*     See Filesystem Dialogs Readme.txt for details                                                                            *
//*                                                                                                                              *
//* Two licenses are available if you like this library:                                                                         *
//* Shareware License: 100 Euros                                                                                                 *
//* Commercial License: 950 Euros                                                                                                *
//*                                                                                                                              *
//*     http://www.3delite.hu/Object Pascal Developer Resources/filesystemdialogs.html                                           *
//*                                                                                                                              *
//* This unit uses the ID3v2 format (a bare Tag file) to store thumbnails of files inside ID3v2.                                 *
//* The structure of the format is: a first frame of "CATA" that stores the catalog of thumbnails in a simple string list format *
//* for example:                                                                                                                 *
//* D:\card.jpg //* Filename with full path                                                                                      *
//* 41639       //* File size - for simplicity the file size is checked weather the file had undergone changes                   *
//* 300         //* Width of the original file, -1 for non image files                                                           *
//* 250         //* Height of the original file, -1 for non image files                                                          *
//* D:\ace.png 	//* The following files...                                                                                       *
//* ...                                                                                                                          *
//* And an "unlimited" number of "THUM" frames containing thumbnail files, in most cases .jpgs that must be in the same order    *
//* as given in the "CATA" frame, that is in the catalog.                                                                        *
//* An "unlimited" number of "SESC" frames containing sample data cache for the audio streams. The SESC frame is compressed, the *
//* THUM is not (as it is .jpg), but could be.                                                                                   *
//* The file is in most cases called: "Thumbs Filesystem Dialogs.db"                                                             *
//* If Filesystem Dialogs is configured to use high quality resample filter, then the quality of image thumbnail is better       *
//* then in Thumbs.db.                                                                                                           *
//* Because full path is stored the database can be everywhere (per folder, 1 global database, etc.).                            *
//* Feel free to add more thumbnail formats (frames).                                                                            *
//*                                                                                                                              *
//* If you have any questions or enquiries please mail: 3delite@3delite.hu                                                       *
//*                                                                                                                              *
//* Good coding! :)                                                                                                              *
//* 3delite                                                                                                                      *
//********************************************************************************************************************************

unit ThumbDatabase;

interface

Uses Classes, SysUtils, Helper3delite, Windows, ID3v2LibraryDefs
    {$IFDEF _DEBUG}
    , FilesystemDialogsDefs
    {$ENDIF}
    //, Dialogs
    ;

const
  FILE_THUMBNAILCACHE       = 'Thumbs Filesystem Dialogs.db';
  FILE_THUMBNAILCACHE_WIN   = 'Thumbs.db';
  FILENAME_DEBUG            = 'C:\Filesystem Dialogs debug.log';

type
  TThumbDatabase = class(TComponent)
  public
    fDatabaseFileName: String;
    fTag: Pointer;
    fEntries: TStrings;
    fThumbCount: Int64;
    fChanged: Boolean;
    Constructor Create(AOwner: TComponent); override;
    Destructor Destroy; override;
    function AddThumb(FullFileName: String; Image: TStream; Width, Height: Int64; FrameID: AnsiString; Compress: Boolean; ForceUpdate: Boolean): Boolean;
    function GetThumb(FileName: String; var Image: TStream; var Width, Height: Int64; var FrameID: AnsiString): Boolean;
    function GetThumbInfo(FileName: String; var Width, Height: Int64; var FrameID: AnsiString): Boolean;
    function SetThumb(FullFileName: String; Image: TStream; Width, Height: Int64; Force: Boolean): Boolean;
    function GetThumbIndex(FullFileNameWithPath: String): Int64;
    function Load(FileName: String): Boolean;
    function Save(FileName: String): Boolean;
  published
  end;

implementation

Constructor TThumbDatabase.Create(AOwner: TComponent);
begin
    inherited Create(AOwner);
    fTag := ID3v2_Create;
    fEntries := TStringList.Create;
    fThumbCount := 0;
    fDatabaseFileName := '';
    fChanged := False;
end;

Destructor TThumbDatabase.Destroy;
begin
    ID3v2_Free(fTag);
    fEntries.Free;
    inherited Destroy;
end;

function TThumbDatabase.GetThumbIndex(FullFileNameWithPath: String): Int64;
var
  i: Integer;
  FileName: String;
  CompareFileName: String;
begin
    Result := 0;
    if fEntries.Count = 0
        then Exit;
    try
        CompareFileName := AnsiUpperCase(ExtractFileName(FullFileNameWithPath));
        for i := 0 to fEntries.Count - 1 do begin
            try
                FileName := fEntries[i];
                FileName := ExtractFileName(FileName);
                if AnsiUpperCase(FileName) = CompareFileName then begin
                    if StrToInt(fEntries[i + 1])  = NGetFileSize(FullFileNameWithPath) then begin
                        Result := (i div 4) + 1;
                        Exit;
                    end;
                end;
            except
                Result := 0;
            end;
        end;
    finally
        //*
    end;
end;

function TThumbDatabase.AddThumb(FullFileName: String; Image: TStream; Width, Height: Int64; FrameID: AnsiString; Compress: Boolean; ForceUpdate: Boolean): Boolean;
var
  FrameIndex: Cardinal;
  i: Integer;
  FileName: String;
  tempIntv: Integer;
  {$IFDEF _DEBUG_SPEED}
  PerfFreq, FreqCountS, FreqCountE, CPUTime: Int64;
  {$ENDIF}
begin

    {$IFDEF _DEBUG_SPEED}
    QueryPerformanceFrequency(PerfFreq);
    QueryPerformanceCounter(FreqCountS);
    {$ENDIF}
    try
        try
            Result := False;
            FrameIndex := 0;
            for i := 0 to fEntries.Count - 1 do begin
                FileName := fEntries[i];
                FileName := ExtractFileName(FileName);
                if AnsiUpperCase(FileName) =  AnsiUpperCase(ExtractFileName(FullFileName)) then begin
                    FrameIndex := (i div 4) + 1;
                    Break;
                end;
            end;
            //* If found replace (if size changed or ForceUpdate)
            if FrameIndex > 0 then begin
                Result := SetThumb(FullFileName, Image, Width, Height, ForceUpdate);
                Exit;
            end;
            //* Else add a new thumb
            FrameIndex := ID3v2_CreateFrame(fTag, PANSIChar(FrameID));
            ID3v2_SetFrameData(fTag, FrameIndex, Image);
            fEntries.Append(FullFileName);
            fEntries.Append(IntToStr(NGetFileSize(FullFileName)));
            fEntries.Append(IntToStr(Width));
            fEntries.Append(IntToStr(Height));
            fThumbCount := fEntries.Count div 4;
            //* Compress frame (3 - max. compression)
            if Compress then begin
                if ID3v2_CompressFrame(fTag, FrameIndex)
                    then Result := True;
            end else Result := True;
            if Result
                then fChanged := True;
        except
            //*
        end;
    finally
        {$IFDEF _DEBUG_SPEED}
        QueryPerformanceCounter(FreqCountE);
        WriteText2File(FILENAME_DEBUG, 'TThumbDatabase.AddThumb(): ' + IntToStr(Round((FreqCountE - FreqCountS) / PerfFreq * 1000)) + ' ms');
   (*     QueryPerformanceFrequency(PerfFreq);
        QueryPerformanceCounter(FreqCountS);
        QueryPerformanceCounter(FreqCountE);
        WriteText2File(FILENAME_DEBUG, 'TThumbDatabase.AddThumb(Save): ' + IntToStr(Round((FreqCountE - FreqCountS) / PerfFreq * 1000)) + ' ms');
  *)   {$ENDIF}
    end;
end;

function TThumbDatabase.GetThumb(FileName: String; var Image: TStream; var Width, Height: Int64; var FrameID: AnsiString): Boolean;
var
  Index: Int64;
  Error: Boolean;
begin
    try
        Result := False;
        Image.Size := 0;
        Index := GetThumbIndex(FileName);
        if Index = 0
            then Exit;
        Inc(Index);
        //* Sync and decompress
        Error := ID3v2_DeCompressFrame(fTag, Index);
        ID3v2_GetFrameData(fTag, Index, Image);
        FrameID := ID3v2_GetFrameID(fTag, Index);
        Image.Seek(0, soFromBeginning);
        Result := True;
        try
            Width := StrToInt(fEntries[(Index - 2) * 4 + 2]);
        except
            Width := 0;
        end;
        try
            Height := StrToInt(fEntries[(Index - 2) * 4 + 3]);
        except
            Width := 0;
        end;
    except
        Image.Size := 0;
    end;
end;

function TThumbDatabase.GetThumbInfo(FileName: String; var Width, Height: Int64; var FrameID: AnsiString): Boolean;
var
  Index: Int64;
begin
    try
        Result := False;
        { TODO 2 -o3delite -cThumb database : Query by FrameID }
        Index := GetThumbIndex(FileName);
        if Index = 0
            then Exit;
        Inc(Index);
        //* Sync
        {
        if fTag.Frames.isUnsynced(Index)
            then fTag.Frames.sync(Index);
        //* Uncompress
        if fTag.Frames.isCompressed(Index)
            then fTag.Frames.decompress(Index);
        fTag.getFrameData(Index, Image);
        }
        FrameID := ID3v2_GetFrameID(fTag, Index);
        if Length(FrameID) > 4
            then FrameID := Copy(FrameID, 0, 4);
        //Image.Seek(0, soFromBeginning);
        Result := True;
        try
            Width := StrToInt(fEntries[(Index - 2) * 4 + 2]);
        except
            Width := 0;
        end;
        try
            Height := StrToInt(fEntries[(Index - 2) * 4 + 3]);
        except
            Width := 0;
        end;
    except
        //*
    end;
end;

function TThumbDatabase.SetThumb(FullFileName: String; Image: TStream; Width, Height: Int64; Force: Boolean): Boolean;
var
  FrameIndex: Cardinal;
  i: Integer;
  FileName: String;
  FileSize: Int64;
begin
    try
        Result := False;
        FrameIndex := 0;
        for i := 0 to fEntries.Count - 1 do begin
            FileName := fEntries[i];
            FileName := ExtractFileName(FileName);
            if AnsiUpperCase(FileName) =  AnsiUpperCase(ExtractFileName(FullFileName)) then begin
                FrameIndex := (i div 4) + 1;
                Break;
            end;
        end;
        if FrameIndex = 0
            then Exit;
        try
            FileSize := StrToInt(fEntries[(FrameIndex * 4) + 1]);
        except
            FileSize := 0;
        end;
        if FileSize = NGetFileSize(FullFileName)
            then if NOT Force
                then Exit;
        ID3v2_SetFrameData(fTag, FrameIndex + 1, Image);
        fEntries[(FrameIndex * 4)] := FullFileName;
        fEntries[(FrameIndex * 4) + 1] := IntToStr(NGetFileSize(FullFileName));
        fEntries[(FrameIndex * 4) + 2] := IntToStr(Width);
        fEntries[(FrameIndex * 4) + 3] := IntToStr(Height);
        Result := True;
        fChanged := True;
    except
        //*
    end;
end;

function TThumbDatabase.Load(FileName: String): Boolean;
var
  CATAIndex: Int64;
  Data: TStream;
  TagStream: TStream;
  Enc: TEncoding;
begin
    fEntries.Clear;
    fThumbCount := 0;
    if FileExists(FileName) then begin
        TagStream := TFileStream.Create(FileName, fmOpenRead);
        ID3v2_LoadFromStream(fTag, TagStream);
        FreeAndNil(TagStream);
    end;
    CATAIndex := ID3v2_CheckFrameExists(fTag, 'CATA');
    if CATAIndex = 0
        then CATAIndex := ID3v2_CreateFrame(fTag, 'CATA');
    Data := TMemoryStream.Create;
    try
        ID3v2_GetFrameData(fTag, CATAIndex, Data);
        Data.Seek(0, soFromBeginning);
        Enc := TUnicodeEncoding.Create;
        fEntries.LoadFromStream(Data, Enc.Unicode);
        FreeAndNil(Enc);
        {$IFDEF _DEBUG}
        //WriteText2File(FILENAME_DEBUG, 'TThumbDatabase.Load(' + FileName + ') listing contents:' + (#13#10) + fEntries.Text);
        {$ENDIF}
        fThumbCount := fEntries.Count div 4;
    finally
        Data.Free;
    end;
    fDatabaseFileName := FileName;
    Result := True;
    fChanged := False;
end;

function TThumbDatabase.Save(FileName: String): Boolean;
var
  CATAIndex: Int64;
  Data: TStream;
  Error: Word;
  Attribs: Cardinal;
  TagStream: TStream;
  Enc: TEncoding;
begin
    Error := 0;
    Data := nil;
    if fEntries.Count > 0 then begin
        CATAIndex := ID3v2_CheckFrameExists(fTag, 'CATA');
        if CATAIndex = 0
            then Exit;
        Data := TMemoryStream.Create;
        Enc := TUnicodeEncoding.Create;
        fEntries.SaveToStream(Data, Enc.Unicode);
        try
            ID3v2_SetFrameData(fTag, CATAIndex, Data);
            //fEntries.LoadFromStream(Data);
        finally
            FreeAndNil(Data);
            FreeAndNil(Enc);
        end;
    end;
    try
        Attribs := 0;
        if FileExists(FileName) then begin
            Attribs := Attribs + GetFileAttributes(PChar(FileName));
            if Attribs = $FFFFFFFF
                then Attribs := 0
                else SetFileAttributes(PChar(FileName), Attribs - FILE_ATTRIBUTE_HIDDEN);
        end else CreateDumbFile(FileName);
        TagStream := TFileStream.Create(FileName, fmOpenReadWrite OR fmShareDenyNone);
        Error := ID3v2_SaveToStream(fTag, TagStream);
        FreeAndNil(TagStream);
        SetFileAttributes(PChar(FileName), Attribs OR FILE_ATTRIBUTE_HIDDEN);
    except
        //*
    end;
    if Error = 0 then begin
        Result := True;
        fDatabaseFileName := FileName;
        fChanged := False;
    end;
end;

end.
